/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2017 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * Evaluating an expression tree in parallel 
 */

define eval_expr(Tree) = {
    if is_integer(Tree) then empty
    else {
	eval_args(Tree);
	apply(Tree[op],Tree[lt],Tree[rt],Tree)
    }
}.

/* Evaluate both arguments of an expression in parallel */
define eval_args(Tree) = {
    halt(is_integer(Tree[lt]) and is_integer(Tree[rt]))
    and stable(struct(Tree))
    and stable(Tree[op])
    and eval_subexpr(Tree[lt])
    and eval_subexpr(Tree[rt])
}.

/* Evaluate a subexpression */
define eval_subexpr(Tree) = {
    eval_expr(Tree);
    stable Tree
}.

define eval_test(t) = {
    exists Tree : {
	Tree=t
	and eval_expr(Tree)
        and always format("Tree=%t\n",Tree)
    }
}.

define apply(fn,A,B,C) = {
    if fn="+" then add(A,B,C)
    else if fn="*" then mult(A,B,C)
    else false
}.

define add(A,B,C) = { skip and C <- A+B }.
define mult(A,B,C) = { skip and C <- A*B }.
/*define mult(A,B,C) = { len(3) and keep_stable(C) and C <- A*B }.*/
/*load "mult".*/

/*define mult(A,B,C) = exists a,b,M,N,Done : {
    Done = true and
    a = A and
    b = B and
    M = 0 and
    N = 0 and
    mult_imp(A,B,M,N,Done) and
    next{halt Done} and
    keep_stable(C) and
    keep(C=["*",M,N]) and
    fin{C=M}
}.
*/

define op=0.
define lt=1.
define rt=2.

define eval_test1 = 1.

define eval_test2 = ["+",1,2].

define eval_test3 = ["*",["+",1,["*",2,3]],["+",4,5]].

define eval_test4 = ["+",
			["*",
			    ["+",1,1],
			    ["+",1,1]],
			["*",
			    ["+",1,1],
			    ["+",1,1]]].

define eval_test5 = ["+",["*",eval_test3,eval_test3],eval_test3].

define eval_test6 = 
  ["+",["*",["+",1,2],3],["+",["*",4,5],["+",6,7]]].

load "misc".
