/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 2014-2017 Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

load "s7-1200-basic".
load "lib-conv".

set outfile="/dev/null".

set print_states=false.
    
/* the following ladder logic 
        In0           In|In|       Out0
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    |   In0           In|In|       Out1       |
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    |   In0           In|In|        Out2      |
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    |   In0           In|In|        Out3      |
    |---| |---| | ... | |-----------( )-------|
    |                 |                       |
    .
    .
    .
    |   In0           In|In|        Out|Out|  |
    
    is translated into Tempura using the constructs defined in s7-1200-basic.t
    
 */
define plc(In,Out) = {
  /* let's output the value of In and Out in each state */
  /*always ( format("In=%t, Out=%t\n",In,Out) )  and*/
    for i<|Out| do {
     /* rung i */
            COIL(AND_IL(In),Out[i]) and skip  and
     forall k<|Out| : { if k ~= i then (next Out[k]) = Out[k] }
    }
}.
    
    


/* X : used to store date read from external file
   I: input I.0 .. I.(ni*8)-1 each 1 bit
   Q: output Q.0 .. Q.(nq*8)-1 each 1 bit
   PL: expected output PL.0 .. PL.(nq*8)-1 each 1 bit 
*/   
define test_plc(ni,nq) = { 
   exists X,I,Q,PL : {
    {set infile="S71200-test.TXT" and empty};
    {   
    /* read the first data element from the file */	   
    input X  and
    /* ni*8 inputs and nq*8 outputs */               
    list(I,ni*8) and stable(struct(I)) and
    list(Q,nq*8) and stable(struct(Q)) and              
    list(PL,nq*8) and stable(struct(PL)) and
    (forall i<nq*8 : {   
    /* Qi and PLi are initially zero */
        Q[i]=0 and
       PL[i]=0
    } ) and
    (forall i<ni*8 : {
    /* Ii is initially zero */
        I[i]=0
    } ) and
    /* as long the input is not zero we continue otherwise we stop */
    while (X ~= 0) do {
       /* we discard the first input as it is MK: data */
            {skip and (next I) = I and (next Q) = Q and (next PL) = PL };
       /* the second data element is input to the plc and stored in I */
       { input X  and skip and
         if X[0..3]="DI:" then {
                 (forall i<ni :{  (next I[i*8..8*(i+1)]) = inttobits(hexbinbyte(X[3+i*3],X[4+3*i]))})
         } and
               (next Q) = Q  and (next PL) = PL
       };
       /* the third data element is output from the plc and stored in Q */
       { input X  and skip and
         if X[0..3]="DO:" then {
                 (forall i<nq :{  (next Q[i*8..8*(i+1)])  = inttobits(hexbinbyte(X[3+i*3],X[4+3*i]))})
         } and
               (next I) = I  and (next PL) = PL 
       }; /* the fourth data element is CT data and discarded */
       { input X and skip and
               (next I) = I and (next Q) = Q and (next PL) = PL 
       }; /* the fifth data element is TI data and discarded */
       { input X  and skip and
               (next I) = I  and (next Q) = Q and (next PL) = PL
       }; /* we feed the input I to our formal specified plc and store
             the output in PL */
       { plc(I,PL) and stable(I) and stable(Q) 
       }; /* we start all over again but also check whether the
             received output Q matches the output PL from the
	     formal specification */
       { input X  and empty and
         forall j<|Q| : {
          if PL[j] ~= Q[j] then {
                  format("Output Q[%d]=%t does not match expected output PL[%d]=%t\n",j,Q[j],j,PL[j])
	  }
	 }   
       }	  
    }
  }
 }   
}.


message("Testing",1,1).
set outfile="/dev/null".
run test_plc(1, 1).
message("Testing",1,2).
set outfile="/dev/null".    
run test_plc(1, 2).
message("Testing",1,4).
set outfile="/dev/null".    
run test_plc(1, 4).    
message("Testing",1,8).
set outfile="/dev/null".
run test_plc(1, 8).
message("Testing",1,16).
set outfile="/dev/null".
run test_plc(1,16).

message("Testing",2,1).
set outfile="/dev/null".
run test_plc(2, 1).
message("Testing",2,2).
set outfile="/dev/null".    
run test_plc(2, 2).
message("Testing",2,4).
set outfile="/dev/null".    
run test_plc(2, 4).    
message("Testing",2,8).
set outfile="/dev/null".
run test_plc(2, 8).
message("Testing",2,16).
set outfile="/dev/null".
run test_plc(2,16).    

message("Testing",4,1).
set outfile="/dev/null".
run test_plc(4, 1).
message("Testing",4,2).
set outfile="/dev/null".    
run test_plc(4, 2).
message("Testing",4,4).
set outfile="/dev/null".    
run test_plc(4, 4).    
message("Testing",4,8).
set outfile="/dev/null".
run test_plc(4, 8).
message("Testing",4,16).
set outfile="/dev/null".
run test_plc(4,16).    

message("Testing",8,1).
set outfile="/dev/null".
run test_plc(8, 1).
message("Testing",8,2).
set outfile="/dev/null".    
run test_plc(8, 2).
message("Testing",8,4).
set outfile="/dev/null".    
run test_plc(8, 4).    
message("Testing",8,8).
set outfile="/dev/null".
run test_plc(8, 8).
message("Testing",8,16).
set outfile="/dev/null".
run test_plc(8,16).

message("Testing",16,1).
set outfile="/dev/null".
run test_plc(16, 1).
message("Testing",16,2).
set outfile="/dev/null".    
run test_plc(16, 2).
message("Testing",16,4).
set outfile="/dev/null".    
run test_plc(16, 4).    
message("Testing",16,8).
set outfile="/dev/null".
run test_plc(16, 8).
message("Testing",16,16).
set outfile="/dev/null".
run test_plc(16,16).
    
set outfile="stdout".    
