/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2017 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-expr.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"

/* The following are from http://xorshift.di.unimi.it/ */
/* slightly adapted for Tempura                        */
uint64_t splitmix64(uint64_t x) {
	uint64_t z = (x += UINT64_C(0x9E3779B97F4A7C15));
	z = (z ^ (z >> 30)) * UINT64_C(0xBF58476D1CE4E5B9);
	z = (z ^ (z >> 27)) * UINT64_C(0x94D049BB133111EB);
	return z ^ (z >> 31);
}

uint64_t s[2];

void init_xorshift128plus(uint64_t x)
{
    struct timeval tnow;
    uint64_t seed64;
    if (x == 0) {
        gettimeofday(&tnow, NULL);
        seed64 = (uint64_t)tnow.tv_usec;
        if ( seed64 == 0 ) {
            seed64 = UINT64_C(42);
        }
    } else {
        seed64 = x;
    }
    printf("Using random_seed = %" PRId64 "\n",seed64);
    s[0] = splitmix64(seed64);
    s[1] = splitmix64(s[0]);
}

double float_xorshift128plus(void) 
{
    uint64_t x = xorshift128plus();
    double d;
    
    x = UINT64_C(0x3FF) << 52 | x >> 12;
    d = *((double *)&x) - 1.0;

    return d;
}


uint64_t xorshift128plus(void) {
        if (s[0] == 0 && s[1] == 0) {
            init_xorshift128plus(0);
        }
        uint64_t s1 = s[0];
	const uint64_t s0 = s[1];
	s[0] = s0;
	s1 ^= s1 << 23; // a
	s[1] = s1 ^ s0 ^ (s1 >> 18) ^ (s0 >> 5); // b, c
	return s[1] + s0; 
}

/* end of xorshift */

/****************************************\
*                                        *
*  statements and expressions for types  *
*                                        *
\****************************************/


/* Get the type of something. */
/* type e */
void reduce_TYPE(e)
NODEPTR e;
{
    NODEPTR loc =reduce_ref  (node_lnk_1(e));

    if (Ready) {
       if ((Ready = mem_defined1 (loc)))
           result (string_node (strsave (print_name (get_mem (loc)))));
    }
}


/* Check whether e is undefined. */
/* _undefined e */
void reduce_iUNDEFINED(e)
NODEPTR e;
{
     NODEPTR loc=reduce_ref (node_lnk_1(e));
     
     if (DebugLevel>=5) {
         printf("<< FO=%1d, SPF=%1d, PPF=%1d, Rdy=%1d, Rtry=%1d>>\n",ForceOutput,StateProgressFlag,PassProgressFlag,Ready,Retry);
     }
         /*if ((Ready && mem_defined1 (loc)) || ForceOutput) {*/
     if ((Ready &&  mem_defined1 (loc))|| ( ForceOutput)) {
             if (mem_defined1 (loc)){
                 if (val_defined2(get_mem(loc))) {
                     node_mem_isundefined(loc)=0;
                     result(False);
                 }
                 else {
                   node_mem_isundefined(loc)=1;
                   result(True);  
                 }
             }
             else{
                 node_mem_isundefined(loc)=1;
                 result(True);
             }
         } 
             /* }*/
     
    
         Ready = (Ready &&  mem_defined1 (loc)) || (ForceOutput) ;
}

/**************************************\
*                                      *
*  Simple Boolean, arithmetic, string  *
*  and list expressions.               *
*                                      *
\**************************************/

/*Equality and Non-equality. */
/* e1 = e2, e1 ~= e2 */
void reduce_EQ_val(e)
NODEPTR e;
{
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        e2 = reduce_val (node_lnk_2(e));
        if (Ready) {
            if (node_type(e) == EQ)
                result (equal_val (e1, e2) ? True : False);
            else
                result (equal_val (e1, e2) ? False : True);
        }
    }
}


#define __HALF_MAX_SIGNED(type) ((type)1 << (sizeof(type)*8-2))
#define __MAX_SIGNED(type) (__HALF_MAX_SIGNED(type) - 1 + __HALF_MAX_SIGNED(type))
#define __MIN_SIGNED(type) (-1 - __MAX_SIGNED(type))

#define __MIN(type) ((type)-1 < 1?__MIN_SIGNED(type):(type)0)
#define __MAX(type) ((type)~__MIN(type))

#define assigncheck(dest,src) ({ \
  typeof(src) __x=(src); \
  typeof(dest) __y=__x; \
  (__x==__y && ((__x<1) == (__y<1))?(void)((dest)=__y),0:1); \
})

#define add_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MIN(typeof(c))-(__b)<=(__a))?assigncheck(c,__a+__b):1) : \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a+__b):1); \
})

#define sub_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MAX(typeof(c))+(__b)>=(__a))?assigncheck(c,__a-__b):1) : \
    ((__MIN(typeof(c))+(__b)<=(__a))?assigncheck(c,__a-__b):1); \
})


/* buggy one
#define sub_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a-__b):1) : \
                      ^ oops
    ((__MIN(typeof(c))+(__b)<=(__a))?assigncheck(c,__a-__b):1); \
})
*/

/*int mult32(long a,long b,long* c) {
    long long x=(long long)a*b;
    if (x!=(long)x) return 0;
  *c=x;
  return 1;
  }*/


/*#if defined(__LP64__)*/
int mult(int64_t x, int64_t y, int64_t * result)
{
    *result = 0;
    if (x > 0 && y > 0 && x > INT64_MAX / y) return 1;
    if (x < 0 && y > 0 && x < INT64_MIN / y) return 1;
    if (x > 0 && y < 0 && y < INT64_MIN / x) return 1;
    if (x < 0 && y < 0 && (x <= INT64_MIN || y <= INT64_MIN || -x > INT64_MAX / -y))
        return 1;
    *result = x * y;
    return 0;
}

/*#else
int mult(long x, long y, long * result)
{
    *result = 0;
    if (x > 0 && y > 0 && x > INT32_MAX / y) return 1;
    if (x < 0 && y > 0 && x < INT32_MIN / y) return 1;
    if (x > 0 && y < 0 && y < INT32_MIN / x) return 1;
    if (x < 0 && y < 0 && (x <= INT32_MIN || y <= INT32_MIN || -x > INT32_MAX / -y))
        return 1;
    *result = x * y;
    return 0;
}
#endif
*/

int64_t power(i,j)
int64_t i,j;
{
    int64_t p = 1;
    int64_t product;
    while (j-- > 0){
        assert0(mult (p, i, &product) == 0,"integer overflow in power" );
        p = product;
    }
    return (p);
}

#define int_plus(X,Y) int_node (node_ival(X) + node_ival(Y))
#define float_plus(X,Y) float_node(node_flval(X) + node_flval(Y))
#define int_minus(X,Y) int_node (node_ival(X) - node_ival(Y))
#define float_minus(X,Y) float_node(node_flval(X) - node_flval(Y))
#define int_mult(X,Y) int_node (node_ival(X) * node_ival(Y))
#define float_mult(X,Y) float_node(node_flval(X) * node_flval(Y))
#define int_div(X,Y) int_node (node_ival(X) / node_ival(Y))
#define float_div(X,Y) float_node(node_flval(X) / node_flval(Y))
#define int_mod(X,Y) int_node (mod (node_ival(X), node_ival(Y)))
#define float_mod(X,Y) float_node(fmod(node_flval(X),node_flval(Y)))
#define float_atantwo(X,Y) float_node(atan2(node_flval(X),node_flval(Y)))
#define int_exp(X,Y) int_node (power (node_ival(X), node_ival(Y)))
#define float_flpower(X,Y) float_node(pow(node_flval(X),node_flval(Y)))

/*  calculate value */
/* e1 and e2, e1 or e2 */
/* e1 < e2, e1 <= e2, e1 >= e2, e1 > e2 */ 
/* e1 + e2, e1 - e2, e1 * e2, e1 div e2, e1 mod e2, e1 ** e2 */
NODEPTR calculate_val(op, e1, e2)
unsigned op;
NODEPTR e1, e2;
{
    unsigned type = node_type(e1);
    int64_t val;
    int64_t product;
        /*double fl_product;*/

    
    
      type_check (e2, type);
      switch (op) {
      case AND:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e2 : e1);
	default:
	  break;
	}
	break;
      case OR:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e1 : e2);
	default:
	  break;
	}
	break;
      case LT:
          return (compare_val (e1, e2) < 0 ? True : False);
      case LE:
          return (compare_val (e1, e2) <= 0 ? True : False);
      case GE:
          return (compare_val (e1, e2) >= 0 ? True : False);
      case GT:
          return (compare_val (e1, e2) > 0 ? True : False);
      case PLUS:
	switch (type) {
	case INT:
            assert0(add_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in addition");
            return (int_node(val));
        case FLOAT:
            return (float_plus (e1,e2));
	case LIST:
            return (append_list (e1, e2));
	case STRING:
            return (append_string (e1, e2));
	default:
	  break;
	}
	break;
      case MINUS:
	switch (type) {
	case INT:
            assert0(sub_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in subtraction");
	  return (int_node (val));
        case FLOAT:
            return (float_minus (e1,e2));  
	default:
	  break;
	}
	break;
      case MULT:
	switch (type) {
	case INT:
            assert0(mult (node_ival(e1), node_ival(e2), &product) == 0,"integer overflow in multiplication" );
            return (int_node(product));
        case FLOAT:
            return (float_mult (e1,e2));    
	default:
	  break;
	}
	break;  
      case DIV:
	switch (type) {
	case INT:
            assert0(node_ival(e2) !=0, "division by zero is undefined\n"); 
	  return (int_div (e1, e2));
        case FLOAT:
            assert0(node_flval(e2) !=0, "division by zero is undefined\n");
            return (float_div (e1,e2));   
	default:
	  break;
	}
	break;
      case MOD:
	switch (type) {
	case INT:
            assert0(node_ival(e2) !=0, "division by zero is undefined\n");
	  return (int_mod (e1, e2));
        case FLOAT:
            assert0(node_flval(e2) !=0, "division by zero is undefined\n");
	  return (float_mod (e1, e2));  
	default:
	  break;
	}
	break;
      case ATANTWO:
	switch (type) {
        case FLOAT:
	  return (float_atantwo (e1, e2));  
	default:
	  break;
	}
	break;  
      case POWER:
	switch (type) {
	case INT:
            assert0(node_ival(e2) >=0, "power to should be >=0\n");
            product=power(node_ival(e1), node_ival(e2));
	  return (int_node (product));
        case FLOAT:
            assert0(node_flval(e1)!=0 || node_flval(e2) >0, "domain error power to\n");
            assert0(node_flval(e1)>=0, "domain error power to\n");
            return (float_flpower(e1, e2));
	default:
	  break;
	}
	break;
      default:
	break;
      }
      error_break2 ("\"%s\" is not a %s operator",
		    Keyword[op], Keyword[type]);
      return (False);
}


/* Binary expressions: AND, OR, PLUS, MINUS, ... */
/* boolean expressions are speeded up otherwise use calculate_val*/
void reduce_bin_op(e)
NODEPTR e;
{
    NODEPTR calculate_val ();
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e)) {
            case AND:
                e2 = (node_bval(e1) == F) ? e1 : reduce_val (node_lnk_2(e));
                break;
            case OR:
                e2 = (node_bval(e1) == T) ? e1 : reduce_val (node_lnk_2(e));
                break;
            default:
                e2 = reduce_val (node_lnk_2(e));
        }
    }
    if (Ready)
        result (calculate_val (node_type(e), e1, e2));
}

#define int_uminus(X) int_node (- node_ival(X))
#define int_not(X) (node_ival(X) == 0 ? One : Zero)
#define float_uminus(X) float_node (- node_flval(X))
/* unary operators: NOT, UPLUS, UMINUS, ... */
/* not e1, +e1, -e1 */
void reduce_u_op(e)
NODEPTR e;
{
    NODEPTR e1;

    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e1)) {
            case BOOL:
                switch (node_type(e)) {
                    case NOT:
                        result (node_bval(e1) ? False : True);
                        break;
                    default:
                        error_break1 ("\"%s\" is not a Boolean operator",
                                print_name (e));
                }
                break;
            case INT:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (int_uminus (e1));
                        break;
                    case ITOF:
                        result (float_node((double)node_ival(e1)));
                        break;
                    default:
                        error_break1 ("\"%s\" is not an integer operator",
                                   print_name (e));
                }
                break;
             case FLOAT:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (float_uminus (e1));
                        break;
                    case CEIL:
                        result (int_node(ceil(node_flval(e1))));
                        break;
                    case FLOOR:
                        result (int_node(floor(node_flval(e1))));
                        break;
                    case SQRT:
                        result (float_node(sqrt(node_flval(e1))));
                        break;
                    case EXPO:
                        result (float_node(exp(node_flval(e1))));
                        break;
                    case LOG:
                        assert0(node_flval(e1)>0, "log argument should be greater than zero\n");
                        result (float_node(log(node_flval(e1))));
                        break;
                    case LOGTEN:
                        assert0(node_flval(e1)>0, "log10 argument should be greater than zero\n");
                        result (float_node(log10(node_flval(e1))));
                        break;
                    case SIN:
                        result (float_node(sin(node_flval(e1))));
                        break;
                    case COS:
                        result (float_node(cos(node_flval(e1))));
                        break;
                    case TAN:
                        result (float_node(tan(node_flval(e1))));
                        break;
                    case ASIN:
                        assert0(node_flval(e1)>=-1 && node_flval(e1)<=1, "asin argument should be in range [-1,1]\n");
                        result (float_node(asin(node_flval(e1))));
                        break;
                    case ACOS:
                        assert0(node_flval(e1)>=-1 && node_flval(e1)<=1, "acos argument should be in range [-1,1]\n");
                        result (float_node(acos(node_flval(e1))));
                        break;
                    case ATAN:
                        result (float_node(atan(node_flval(e1))));
                        break;
                    case SINH:
                        result (float_node(sinh(node_flval(e1))));
                        break;
                    case COSH:
                        result (float_node(cosh(node_flval(e1))));
                        break;
                    case TANH:
                        result (float_node(tanh(node_flval(e1))));
                        break;
                    case FABS:
                        result (float_node(fabs(node_flval(e1))));
                        break;    
                    default:
                        error_break1 ("\"%s\" is not an floating point operator",
                                   print_name (e));
                }
                break;   
                    /*case STRING:
                switch (node_type(e)) {
                        case SIZE:
                        result (int_node ((int64_t)node_lnk_2(e)));
                        break;
                    default:
                        error_break1 ("\"%s\" is not a string operator",
                                print_name (e));
                }
                break;*/
            default:
                error_break2 ("\"%s\" is not a %s operator",
                        print_name (e), print_name (e1));
        }
    }
}


/* state_number  */
/* Reduces to current state number */
void reduce_STATENUMBER()
{
  if ( StateNrValTime != Clock || StateNrVal == -1 ) {
    StateNrValTime = Clock;
  };
  Ready = T;
  result (int_node (StateNumber));
}


/* random  */
/* Reduces to random integer >=0  */
void reduce_RANDOM()
{
  if ( RandomValTime != Clock || RandomVal == -1 ) {
          /*RandomVal = i4_uniform_0i (&TRandSeed);*/
          /*RandomVal = genrand64_int63();*/
      RandomVal = xorshift128plus();
      RandomValTime = Clock;
  };
  Ready = T;
  result (int_node (RandomVal));
}

/* Random  */
/* Reduces to random integer >=0  */
/* This will  allow for more than 1 random number
   in the same state */
void reduce_RRANDOM()
{
  if ( RRandomValTime != Clock || RRandomVal == -1 ) {
    RRandomValTime = Clock;
  };
  Ready = T;
      /* result (int_node (i4_uniform_0i (&TRRandSeed)));*/
      /*result (int_node (genrand64_int63()));*/
  result (int_node(xorshift128plus()));
  
}

/* frandom  */
/* Reduces to random in [0.0,1.0) */
void reduce_SRANDOM()
{
  if ( SRandomValTime != Clock || SRandomVal == -1 ) {
          /*SRandomVal = r4_uniform_01 (&TSRandSeed);*/
          /*SRandomVal = genrand64_real2();*/
      SRandomVal = float_xorshift128plus();
      SRandomValTime = Clock;
  };
  Ready = T;
  result (float_node (SRandomVal));
}

/* fRandom  */
/* Reduces to random in [0.0,1.0) */
/* This will  allow for more than 1 frandom number
   in the same state */
void reduce_SRRANDOM()
{
  if ( SRRandomValTime != Clock || SRRandomVal == -1 ) {
    SRRandomValTime = Clock;
  };
  Ready = T;
      /* result (float_node (r4_uniform_01 (&TSRRandSeed)));*/
      /*result (float_node ( genrand64_real2() ));*/
  result (float_node(float_xorshift128plus()));
  
}

/********************\
*                    *
*  string operators  *
*                    *
\********************/

/* ascii(e1) */
void reduce_ASCII(e)
NODEPTR e;
{
    char   *strr;
    NODEPTR e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, STRING);
        assert1 (strlen (strr = node_strval(e1)) == 1,
                "cannot get ascii code of string \"%s\" (length not 1)", strr);
        result (int_node (*strr));
    }
}



/**********************\
*                      *
*  pointer operations  *
*                      *
\**********************/
/* ref(loc) */
void reduce_REFER(e)
NODEPTR e;
{
    NODEPTR loc = reduce_ref (node_lnk_1(e));
    if (Ready) {
        assert0 (loc != TempLoc, "cannot reference a value");
        result (loc_node (loc, is_next(loc))); 
    }
}
/* deref p */
void reduce_DEREF(e)
NODEPTR e;
{
    NODEPTR p = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (p, LOC); 
        ref_result (node_loc_tonext(p) ? next_loc (node_loc_ptr(p)) : curr_loc (node_loc_ptr(p)));
    }
}

